/*
 * Copyright (c) 2020 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: Apache-2.0
 */

/**
 * @file
 * @brief Sample app for Audio class
 */

#include <zephyr/kernel.h>
#include <zephyr/devicetree.h>
#include <zephyr/logging/log.h>
#include <zephyr/usb/usb_device.h>
#include <zephyr/usb/class/usb_audio.h>
#include <zephyr/drivers/gpio.h>
#include <zephyr/net/buf.h>
#include <string.h>

#include <nrfx_i2s.h>
#include <hal/nrf_gpio.h>

#define SAMPLE_FREQUENCY    48000
#define SAMPLE_BIT_WIDTH    16
#define BYTES_PER_SAMPLE    sizeof(int16_t)
#define FRAME_SIZE   48
#define NUMBER_OF_CHANNELS 2
#define SAMPLES_PER_BLOCK   (FRAME_SIZE * NUMBER_OF_CHANNELS)
#define INITIAL_BLOCKS      2
#define TIMEOUT             1000
#define CHUNK_SIZE_BYTES  (BYTES_PER_SAMPLE * SAMPLES_PER_BLOCK)

LOG_MODULE_REGISTER(main, LOG_LEVEL_INF);

static const struct device *const mic_dev = DEVICE_DT_GET_ONE(usb_audio_mic);

#define abs(x) (x > 0 ? x : -x)

#define I2S_SCK_M NRF_GPIO_PIN_MAP(0,26)
#define I2S_LRCK_M NRF_GPIO_PIN_MAP(0,27)
#define I2S_SDOUT NRF_I2S_PIN_NOT_CONNECTED
#define I2S_SDIN NRF_GPIO_PIN_MAP(1,8)

#define I2S_MCK_M NRF_I2S_PIN_NOT_CONNECTED

NET_BUF_POOL_DEFINE(netbuf_pool, 50, CHUNK_SIZE_BYTES, 0, NULL);

bool write_to_usb(struct net_buf *buffer, size_t size){
	int ret;

	if (!buffer || !size) {
		LOG_ERR("BUFF NULL!");
		/* This should never happen */
		return 1;
	}

	/* Check if OUT device buffer can be used for IN device */
	if (size == usb_audio_get_in_frame_size(mic_dev)) {
		ret = usb_audio_send(mic_dev, buffer, size);
		if(ret){
			net_buf_unref(buffer);
			return 1;
		}
	} else {
		LOG_ERR("In frame not equal to size. Found %d, expected %d", usb_audio_get_in_frame_size(mic_dev), size);
		net_buf_unref(buffer);
		return 1;
	}

	return 0;
}

static void feature_update(const struct device *dev,
			   const struct usb_audio_fu_evt *evt)
{
	int16_t volume = 0;

	LOG_DBG("Control selector %d for channel %d updated",
		evt->cs, evt->channel);
	switch (evt->cs) {
	case USB_AUDIO_FU_MUTE_CONTROL:
		break;
	case USB_AUDIO_FU_VOLUME_CONTROL:
		volume = *((int16_t *)(evt->val));
		LOG_INF("set volume: %d", volume);
		break;
	default:
		break;
	}
}

static const struct usb_audio_ops mic_ops = {
	.feature_update_cb = feature_update,
};


ISR_DIRECT_DECLARE(i2s_isr){
	// data_ready_flag = false;
	nrfx_i2s_0_irq_handler();
	ISR_DIRECT_PM(); // PM done after servicing interrupt for best latency
	return 1;
}

int buf_idx=0;
int16_t audio_buf[2 * SAMPLES_PER_BLOCK];
nrfx_i2s_buffers_t i2s_buffers;
nrfx_i2s_t i2s_inst = NRFX_I2S_INSTANCE(0);

void i2s_handler(nrfx_i2s_buffers_t const * p_released, uint32_t status){
	if(status & NRFX_I2S_STATUS_NEXT_BUFFERS_NEEDED){

		struct net_buf *buf = net_buf_alloc(&netbuf_pool, K_NO_WAIT);
		if(buf){
			int16_t* rx_aud = (int16_t*) p_released->p_rx_buffer;
			
			for(int i = 0; i < SAMPLES_PER_BLOCK; i+=2){
				if(i >= 2){
					const int16_t max_cnt = 2304 * 5; // From sender
					if( (max_cnt + rx_aud[i] - rx_aud[i-2]) % (max_cnt) != 5){
						LOG_RAW("%d, %d -> %d\n", i, rx_aud[i-2], rx_aud[i]);
					}
				}
			}

			// Copy data from I2S to USB audio
			for(int i = 0; i < SAMPLES_PER_BLOCK; i++){
				net_buf_add_le16(buf, rx_aud[i]);
			}
			write_to_usb(buf, CHUNK_SIZE_BYTES);
		}

		i2s_buffers.p_rx_buffer = (uint32_t*) (audio_buf + buf_idx * SAMPLES_PER_BLOCK);
		nrfx_i2s_next_buffers_set(&i2s_inst, &i2s_buffers);
		buf_idx = !buf_idx;
	}
}

void init_i2s(){
	IRQ_DIRECT_CONNECT(I2S_IRQn, 1, i2s_isr, 0);
	irq_enable(I2S_IRQn);

	nrfx_i2s_config_t i2s_config = NRFX_I2S_DEFAULT_CONFIG(I2S_SCK_M, I2S_LRCK_M, I2S_MCK_M, I2S_SDOUT, I2S_SDIN);
	i2s_config.irq_priority = 7;
	i2s_config.mode = NRF_I2S_MODE_MASTER;
	i2s_config.format = I2S_CONFIG_FORMAT_FORMAT_I2S;
	i2s_config.sample_width = NRF_I2S_SWIDTH_16BIT;
	i2s_config.channels = NRF_I2S_CHANNELS_STEREO;

	// ~48kHz with master clock = 1523809.5238095238
	i2s_config.mck_setup = NRF_I2S_MCK_32MDIV21; 
	i2s_config.ratio = NRF_I2S_RATIO_32X;
	nrfx_i2s_init(&i2s_inst, &i2s_config, i2s_handler);
}


int16_t _buf[4096];

int main(void){
	int ret;

	LOG_INF("Entered %s", __func__);
	
	// Initialize mic
	if (!device_is_ready(mic_dev)) {
		LOG_ERR("Device USB Microphone is not ready");
		return 0;
	}

	LOG_INF("Found USB Microphone Device");
	usb_audio_register(mic_dev, &mic_ops);
	
	// Initialize I2S
	init_i2s();

	if (ret != 0) {
		LOG_ERR("Failed to enable USB");
		return 0;
	}

	i2s_buffers.p_rx_buffer = (uint32_t*) audio_buf;
	i2s_buffers.p_tx_buffer = NULL;
	nrfx_i2s_start(&i2s_inst, &i2s_buffers, FRAME_SIZE, 0);

	ret = usb_enable(NULL);
	LOG_INF("USB enabled");
	
	while(1){
		k_msleep(1000);
	}

	return 0;
}
